(****** Encoding of A Language for Certified Programming ******)
(******                            August - October 2002 ******)

Require EqNat.
Require Max.
Require Lt.
Require Plus.
Require Arith.
Require Export Coq.Lists.PolyList.
Require Export heap2.

(****** Special tactic to use the fsl hints database ******)
Tactic Definition FAuto := Auto with fsl.
Tactic Definition SplitAllHyps :=  (* splits all conjunctions in hypotheses *)
  Repeat Match Context With [H:?/\? |- ?] -> Inversion_clear H.
Tactic Definition SplitAllExs :=   (* Splits all existentials in hypothesis *)
  Repeat Match Context With [H:(ex ? ?) |- ?] -> Inversion_clear H.


(****** Syntactic Categories & Related Utilities ******)

Syntactic Definition lab := nat.		(* code label *)

Inductive reg : Set :=				(* registers *)
   r0 : reg | r1 : reg | r2 : reg | r3 : reg
 | r4 : reg | r5 : reg | r6 : reg | r7 : reg
 | r8 : reg | r9 : reg | r10 : reg | r11 : reg
 | r12 : reg | r13 : reg | r14 : reg | r15 : reg
 | r16 : reg | r17 : reg | r18 : reg | r19 : reg
 | r20 : reg | r21 : reg | r22 : reg | r23 : reg
 | r24 : reg | r25 : reg | r26 : reg | r27 : reg
 | r28 : reg | r29 : reg | r30 : reg | r31 : reg.

Definition beq_reg :=				(* comp r1 r2, ret bool *)
 [r1,r2:reg]
 Cases r1 r2 of
  r0 r0 => true | r1 r1 => true | r2 r2 => true | r3 r3 => true
 | r4 r4 => true | r5 r5 => true | r6 r6 => true | r7 r7 => true
 | r8 r8 => true | r9 r9 => true | r10 r10 => true | r11 r11 => true
 | r12 r12 => true | r13 r13 => true | r14 r14 => true | r15 r15 => true
 | r16 r16 => true | r17 r17 => true | r18 r18 => true | r19 r19 => true
 | r20 r20 => true | r21 r21 => true | r22 r22 => true | r23 r23 => true
 | r24 r24 => true | r25 r25 => true | r26 r26 => true | r27 r27 => true
 | r28 r28 => true | r29 r29 => true | r30 r30 => true | r31 r31 => true
 | _ _ => false
 end.

Inductive comm : Set :=				(* commands *)
   add : reg -> reg -> reg -> comm
 | addi : reg -> reg -> int -> comm
 | sub : reg -> reg -> reg -> comm
 | subi : reg -> reg -> int -> comm
 | mov : reg -> reg -> comm
 | movi : reg -> int -> comm
 | bgt : reg -> reg -> lab -> comm
 | bgti : reg -> int -> lab -> comm
 | alloc : reg -> reg -> comm
 | ld : reg -> reg -> nat -> comm
 | st : reg -> nat -> reg -> comm
 | free : reg -> comm.

Inductive iseq : Set :=				(* instruction sequence *)
  seq : comm -> iseq -> iseq
 | jd : lab -> iseq
 | jmp : reg -> iseq.

Inductive rfile : Set :=			(* register file *)
   emptyR : rfile
 | consR : rfile -> reg -> int -> rfile.

Fixpoint lookupR[rf:rfile] : reg -> int :=	(* look up reg file *)
 [r:reg]
 Cases rf of
  emptyR => O
 | (consR rf' r' i) => (if (beq_reg r r') then i
                        else (lookupR rf' r))
 end.

Definition updateR : rfile -> reg -> int -> rfile := (* update reg file *)
 [rf:rfile; r:reg; i:int]
 (consR rf r i).

Fixpoint hwriteN[n:nat] : heap -> addr -> (nat -> int) -> heap :=
  (* write with multiple elements *)
 [h:heap; i:addr; l:(nat->int)]
 Cases n of
  O => h
 | (S m) => (hwriteN m (hwrite h i (l O)) (S i) ([x:nat] (l (S x))))
 end.

Syntactic Definition pnil := O.			(* nil ptr *)

Inductive WF_heap : heap -> Prop :=		(* well-formed Heap *)
  wf_h : (h:heap) ((h pnil)=hvnone) -> (WF_heap h).

Syntactic Definition state := (heap * rfile).	(* state *)

Definition getValinH [s:state] : addr -> hvopt := (* get val from H of state *)
 [i:addr]
  let (H, R) = s in (hread H i).

Definition getValinR [s:state] : reg -> int :=	(* get val from R of state *)
 [r:reg]
  let (H, R) = s in (lookupR R r).

Definition removeS[s:state] : addr -> state :=	(* remove elem from H of S *)
 [i:addr]
  let (H, R) = s in ((hfree H i), R).

Syntactic Definition assert := state -> Prop.	(* asserts *)

Inductive codeheap : Set :=			(* code heap *)
 | emptyC : codeheap
 | consC : lab -> iseq -> codeheap -> codeheap.

Inductive codeheapty : Type :=			(* code heap type *)
 | emptyCT : codeheapty
 | consCT : lab -> assert -> codeheapty -> codeheapty.

Inductive lookupC : codeheap -> lab -> iseq -> Prop := (* look up code heap *)
 | lookupC_eq : (ch:codeheap; l:lab; b:iseq)
                 (lookupC (consC l b ch) l b)
 | lookupC_neq : (ch:codeheap; l,l':lab; b,b':iseq)
		  (lookupC ch l' b')
		   -> (~l=l')
		   -> (lookupC (consC l b ch) l' b').

Inductive lookupCT : codeheapty -> lab -> assert -> Prop :=
  (* look up code heap type *)
 | lookupCT_eq : (ct:codeheapty; l:lab; a:assert)
                 (lookupCT (consCT l a ct) l a)
 | lookupCT_neq : (ct:codeheapty; l,l':lab; a,a':assert)
		  (lookupCT ct l' a')
		   -> (~l=l')
		   -> (lookupCT (consCT l a ct) l' a').

Syntactic Definition prog := (codeheap * state * iseq). (* program *)

(****** Shorthands for Assertions ******)

Definition Aand[a1,a2:assert] : assert :=	(* and *)
 [s:state]
 (a1 s) /\ (a2 s).

Definition Aor[a1,a2:assert] : assert :=	(* or *)
 [s:state]
 (a1 s) \/ (a2 s).

Definition emp : assert :=			(* empty heap *)
 [s:state]
 ((i:addr) ((getValinH s i) = hvnone)).

Definition pto[i:addr;n:int] : assert :=	(* singleton heap *)
 [s:state]
 ((getValinH s i) = (hvsome n)
  /\ ((j:addr) (~i=j) -> ((getValinH s j) = hvnone))).

Definition star[a1,a2:assert] : assert :=	(* separating conjunction *)
 [s:state]
  let (H, R) = s in
 (EX H1:heap |
   (EX H2:heap |
     (hdisj H1 H2) /\ (heap_eq (happend H1 H2) H)
      /\ (a1 (H1, R)) /\ (a2 (H2, R)))).

Definition star1 : addr -> int -> assert -> assert := (* (i->n)*a *)
 [i:addr][n:int][a:assert][s:state]
 ((getValinH s i) = (hvsome n) /\ (a (removeS s i))).

Definition simp[a1,a2:assert] : assert :=	(* separating implication *)
 [s:state]
  let (H, R) = s in
 ((H':heap)
   (hdisj H H') -> (a1 (H', R)) -> (a2 ((happend H H'), R))).

Definition ptoany[a:addr] : assert :=	(* singleton heap w. unknown content *)
 [s:state]
 (EX x:int | ((pto a x) s)).

Fixpoint ptolist[a:addr; l:(list int)] : assert := (* mult. elem. heap *)
 Cases l of
  nil => emp
 | (cons i l') => (star1 a i (ptolist (S a) l'))
 end.

(****** Inference Rules ******)

Inductive Infer : codeheapty -> assert -> iseq -> Prop :=
  (* pure rules *)
  I_add : (ct:codeheapty; a,a':assert; rd,rs,rt:reg; b:iseq)
          ((H:heap; R:rfile)
           (a (H, R))
	   -> (a' (H, (updateR R rd (plus (lookupR R rs) (lookupR R rt))))))
	  -> (Infer ct a' b)
	  -> (Infer ct a (seq (add rd rs rt) b))
 | I_addi : (ct:codeheapty; a,a':assert; rd,rs:reg; i:int; b:iseq)
	    ((H:heap; R:rfile)
	     (a (H, R))
	     -> (a' (H, (updateR R rd (plus i (lookupR R rs))))))
	    -> (Infer ct a' b)
	    -> (Infer ct a (seq (addi rd rs i) b))
 | I_sub : (ct:codeheapty; a,a':assert; rd,rs,rt:reg; b:iseq)
          ((H:heap; R:rfile)
           (a (H, R))
	   -> (a' (H, (updateR R rd (minus (lookupR R rs) (lookupR R rt))))))
	  -> (Infer ct a' b)
	  -> (Infer ct a (seq (sub rd rs rt) b))
 | I_subi : (ct:codeheapty; a,a':assert; rd,rs:reg; i:int; b:iseq)
	    ((H:heap; R:rfile)
	     (a (H, R))
	     -> (a' (H, (updateR R rd (minus (lookupR R rs) i)))))
	    -> (Infer ct a' b)
	    -> (Infer ct a (seq (subi rd rs i) b))
 | I_mov : (ct:codeheapty; a,a':assert; rd,rs:reg; b:iseq)
	   ((H:heap; R:rfile)
	    (a (H, R))
	    -> (a' (H, (updateR R rd (lookupR R rs)))))
	   -> (Infer ct a' b)
	   -> (Infer ct a (seq (mov rd rs) b))
 | I_movi : (ct:codeheapty; a,a':assert; rd:reg; i:int; b:iseq)
	    ((H:heap; R:rfile)
	     (a (H, R))
	     -> (a' (H, (updateR R rd i))))
	    -> (Infer ct a' b)
	    -> (Infer ct a (seq (movi rd i) b))
 | I_bgt : (ct:codeheapty; a,a',a1:assert; rs,rt:reg; f:lab; b:iseq)
	   (lookupCT ct f a1)
	    -> ((H:heap; R:rfile)
	        (le (lookupR R rs) (lookupR R rt))
		-> (a (H, R))
		-> (a' (H, R)))
	    -> ((H:heap; R:rfile)
	        (gt (lookupR R rs) (lookupR R rt))
	        -> (a (H, R))
	        -> (a1 (H, R)))
	    -> (Infer ct a' b)
	    -> (Infer ct a (seq (bgt rs rt f) b))
 | I_bgti : (ct:codeheapty; a,a',a1:assert; rs:reg; i:int; f:lab; b:iseq)
	   (lookupCT ct f a1)
	    -> ((H:heap; R:rfile)
	        (le (lookupR R rs) i)
		-> (a (H, R))
		-> (a' (H, R)))
	    -> ((H:heap; R:rfile)
	        (gt (lookupR R rs) i)
	        -> (a (H, R))
	        -> (a1 (H, R)))
	    -> (Infer ct a' b)
	    -> (Infer ct a (seq (bgti rs i f) b))
 | I_jd : (ct:codeheapty; a,a1:assert; f:lab)
          (lookupCT ct f a1)
	  -> ((s:state) (a s) -> (a1 s))
	  -> (Infer ct a (jd f))
 | I_jmp : (ct:codeheapty; a:assert; r:reg)
	   ((H:heap; R:rfile)
	    (a (H, R)) -> (exT assert ([a1:assert]
	     (lookupCT ct (lookupR R r) a1)
	     /\ (a1 (H, R)))))
           -> (Infer ct a (jmp r))
  (* impure rules *)
 | I_alloc : (ct:codeheapty; a,a':assert; rs,rd:reg; b:iseq)
	     ((H:heap; R:rfile; l:addr; nl:(nat->int))
	      (a (H, R))
              -> (gt l O)
	      -> ((i:nat) (gt (lookupR R rs) i)
	          -> (hnotindom (plus i l) H))
	      -> (a' ((hwriteN (lookupR R rs) H l nl),
	              (updateR R rd l))))
	     -> (Infer ct a' b)
	     -> (Infer ct a (seq (alloc rd rs) b))
 | I_ld : (ct:codeheapty; a,a':assert; rd,rs:reg; i:nat; b:iseq)
	  ((H:heap; R:rfile)
	   (a (H, R))
	   -> (EX n | ((hread H (plus i (lookupR R rs))) = (hvsome n))
	           /\ (a' (H, (updateR R rd n)))))
	  -> (Infer ct a' b)
	  -> (Infer ct a (seq (ld rd rs i) b))
 | I_st : (ct:codeheapty; a,a':assert; rd,rs:reg; i:nat; b:iseq)
	  ((H:heap; R:rfile)
	   (a (H, R))
	   -> ((hindom (plus i (lookupR R rd)) H)
	    /\ (a' ((hwrite H (plus i (lookupR R rd)) (lookupR R rs)), R))))
	  -> (Infer ct a' b)
	  -> (Infer ct a (seq (st rd i rs) b))
 | I_free : (ct:codeheapty; a,a':assert; rs:reg; b:iseq)
	    ((H:heap; R:rfile)
	     (a (H, R))
	     -> ((hindom (lookupR R rs) H)
	      /\ (a' ((hfree H (lookupR R rs)), R))))
	    -> (Infer ct a' b)
	    -> (Infer ct a (seq (free rs) b)).

Inductive WF_CH_0 : codeheapty -> codeheapty -> codeheap -> Prop :=
 | wf_c_emp : (ct0:codeheapty) (WF_CH_0 ct0 emptyCT emptyC)
 | wf_c_cons : (ct0,ct:codeheapty; ch:codeheap; f:lab; a:assert; b:iseq)
               (Infer ct0 a b)
               -> (WF_CH_0 ct0 ct ch)
               -> (WF_CH_0 ct0 (consCT f a ct) (consC f b ch)).

Inductive WF_CH : codeheapty -> codeheap -> Prop :=
 | wf_c : (ct:codeheapty; ch:codeheap)
          (WF_CH_0 ct ct ch)
	  -> (WF_CH ct ch).

Inductive InferP : codeheapty -> assert -> prog -> Prop :=
  I_prog : (ct:codeheapty; a:assert; ch:codeheap; s:state; b:iseq)
           (WF_CH ct ch)
	   -> (Infer ct a b)
	   -> (a s)
           -> (InferP ct a (ch, (s, b))).

(****** Lemmas on Utility Functions ******)

(* arithmetic lemmas *)
Lemma neq_Sn_n : (n:nat) (~(S n)=n).
Proof.
Induction n; [Discriminate | Intros; Apply not_eq_S; Assumption].
Save.
Hints Resolve neq_Sn_n : fsl.

Lemma lt_neq : (n,m:nat) (lt n m) -> ~n=m.
Proof.
Double Induction 1 2; [Intro H; Inversion H | Intros; Auto | Intros; Auto | Intros; Apply not_eq_S; Apply H0; Apply lt_S_n; Auto].
Save.
Hints Resolve lt_neq : fsl.


Lemma beq_nat_sym : (n,m:nat) ((beq_nat n m) = (beq_nat m n)).
Proof.
Double Induction 1 2; Auto; Intros; Compute; Auto.
Save.

Lemma not_beq_Sn_n : (n:nat) ((beq_nat (S n) n)=false).
Proof.
Intros; Apply sym_eq; FAuto.
Save.
Hints Resolve not_beq_Sn_n : fsl.


Lemma not_beq_nat_big_small : (m,n:nat)
 (lt n m) -> ((beq_nat m n) = false).
Proof.
Intros; Apply sym_eq; Apply beq_nat_neq; Apply sym_not_eq; FAuto.
Save.
Hints Resolve not_beq_nat_big_small : fsl.

Lemma not_beq_nat_small_big : (m,n:nat)
 (lt n m) -> ((beq_nat n m) = false).
Proof.
Intros; Replace (beq_nat n m) with (beq_nat m n); [FAuto | Apply beq_nat_sym].
Save.
Hints Resolve not_beq_nat_small_big : fsl.


(* heap lemmas *)

Lemma h_pnil_notindom : (H:heap)
 (WF_heap H) -> (hnotindom pnil H).
Proof.
Intros H H0; Inversion H0; Apply notindom_none; Assumption.
Save.
Hints Resolve h_pnil_notindom : fsl.

Lemma good_hwrite : (H:heap; i:addr; n:int)
 (WF_heap H) -> (~i=O) -> (WF_heap (hwrite H i n)).
Proof.
Intros.
Inversion H0; Apply wf_h.
Unfold hwrite.
Replace (beq_nat i O) with false; [Assumption | FAuto].
Save.
Hints Immediate good_hwrite : fsl.

Lemma good_HwriteN : (n:nat; H:heap; a:addr; l:(nat->int))
 (WF_heap H) ->
 ((i:nat)(lt i n) -> (~(plus i a)=O)) ->
 (WF_heap (hwriteN n H a l)).
Proof.
Induction n; Intros.
Unfold hwriteN; Assumption.
Unfold hwriteN; Fold hwriteN.
Apply H.
Apply good_hwrite; Auto.
Replace a with (plus O a).
Apply H2.
Apply lt_O_Sn.
Auto.
Intros.
Replace (plus i (S a)) with (plus (S i) a).
Apply H2.
Apply lt_n_S; Assumption.
Apply plus_Snm_nSm.
Save.

Lemma good_Hfree : (H:heap; i:addr)
 (WF_heap H) -> (WF_heap (hfree H i)).
Proof.
Intros.
Inversion H0; Apply wf_h.
Unfold hfree.
Elim (beq_nat i O); Auto.
Save.

Lemma good_Happend : (H1,H2:heap)
 (hdisj H1 H2) -> 
 (WF_heap H1) -> (WF_heap H2) ->
 (WF_heap (happend H1 H2)).
Proof.
Unfold happend; Intros.
Inversion_clear H; Inversion_clear H0; Inversion_clear H3; Apply wf_h.
Rewrite H; Assumption.
Save.
