(****** heaps -- encoded as a partial function,     ******)
(******          where the ordering does not matter ******)
(******                              September 2002 ******)

Require Export EqNat.

(* syntactic categories *)
Syntactic Definition int := nat.		(* integers *)
Syntactic Definition addr := nat.		(* heap labels *)
Syntactic Definition hval := int.		(* heap values *)
Syntactic Definition hvopt := (option hval).	(* range of heap *)

(* short hands for options *)
Syntactic Definition hvsome := (Some hval).	(* allocated *)
Syntactic Definition hvnone := (None hval).	(* unallocated *)

Definition heap : Set := (addr -> hvopt).	(* heap *)

Inductive heap_eq : heap -> heap -> Prop := 	(* equality of heap *)
  all_same : (h1,h2:heap)
             ((i:addr)(h1 i)=(h2 i)) ->
	     (heap_eq h1 h2).

Definition hwrite[h:heap] : addr -> hval -> heap := (* write to heap *)
 [i:addr; n:hval]
 [j:addr] (if (beq_nat i j) then (hvsome n) else (h j)).

Definition hread[h:heap] : addr -> hvopt :=	(* read from heap *)
 [i:addr] (h i).

Definition hfree[h:heap] : addr -> heap :=	(* free heap cell *)
 [i:addr]
 [j:addr] (if (beq_nat i j) then hvnone else (h j)).

Inductive hnotindom : addr -> heap -> Prop :=	(* cell not in domain *)
  notindom_none : (i:addr; h:heap)
                  ((h i)=hvnone) ->
		  (hnotindom i h).

Inductive hindom : addr -> heap -> Prop :=	(* cell in domain *)
  indom_some : (i:addr; h:heap; n:hval)
               ((h i)=(hvsome n)) ->
	       (hindom i h).

Inductive hdisj : heap -> heap -> Prop :=	(* disjoint domain *)
  disj_notboth : (h1,h2:heap)
                 ((i:addr)(hnotindom i h1) \/ (hnotindom i h2)) ->
		 (hdisj h1 h2).

Definition happend[h1,h2:heap] : heap :=	(* append two heaps *)
  (* precondition is that (disj h1 h2) *)
 [i:addr] (
   Cases (h1 i) of
   | (Some n) => (hvsome n)
   | None => (h2 i)
   end).

Definition hempty : heap :=			(* empty heap *)
 [i:addr] hvnone.

Definition hsingleton : addr -> hval -> heap :=	(* singleton heap *)
 [i:addr; n:hval]
 [j:addr] (if (beq_nat i j) then (hvsome n) else hvnone).

Lemma beq_nat_neq : (i,j:nat)
 (~i=j) -> (false = (beq_nat i j)).
Proof.
Double Induction 1 2.
Intro; Absurd O=O->False; Auto.
Auto.
Auto.
Intros.
Compute; Fold beq_nat; Auto.
Save.

Lemma opt_dec : (A:Set)(o:(option A))
  o=(None A)\/(EX n:A | o=(Some A n)).
Proof.
Induction o.
Intros; Right; Exists a; Trivial.
Left; Trivial.
Save.

Lemma heap_eq_refl : (h:heap)
 (heap_eq h h).
Proof.
Intros; Apply all_same; Trivial.
Save.

Lemma heap_eq_sym : (h1,h2:heap)
 (heap_eq h1 h2) -> (heap_eq h2 h1).
Proof.
Intros.
Inversion H; Apply all_same.
Intro; Cut (h1 i)=(h2 i).
Intro EQ; Rewrite EQ; Trivial.
Apply H0; Assumption.
Save.

Lemma heap_eq_trans : (h1,h2,h3:heap)
 (heap_eq h1 h2) -> (heap_eq h2 h3) -> (heap_eq h1 h3).
Proof.
Intros.
Inversion_clear H; Inversion_clear H0; Apply all_same.
Intro; Replace (h3 i) with (h2 i); Auto.
Save.

Lemma h_subst_write : (h1,h2:heap; i:addr; n:hval)
 (heap_eq h1 h2) -> (heap_eq (hwrite h1 i n) (hwrite h2 i n)).
Proof.
Intros.
Unfold hwrite; Inversion_clear H; Apply all_same.
Intro; Generalize (beq_nat i i0).
Induction b; Auto.
Save.

Lemma h_subst_read : (h1,h2:heap; i:addr)
 (heap_eq h1 h2) -> ((hread h1 i) = (hread h2 i)).
Proof.
Intros.
Unfold hread; Inversion_clear H; Auto.
Save.

Lemma h_subst_free : (h1,h2:heap; i:addr)
 (heap_eq h1 h2) -> (heap_eq (hfree h1 i) (hfree h2 i)).
Proof.
Intros.
Unfold hfree; Inversion_clear H; Apply all_same.
Intro; Generalize (beq_nat i i0).
Induction b; Auto.
Save.

Lemma h_subst_notin : (h1,h2:heap; i:addr)
 (heap_eq h1 h2) -> (hnotindom i h1) -> (hnotindom i h2).
Proof.
Intros.
Inversion_clear H0; Inversion_clear H; Apply notindom_none.
Rewrite <- H1; Apply sym_eq; Apply H0.
Save.

Lemma h_subst_indom : (h1,h2:heap; i:addr)
 (heap_eq h1 h2) -> (hindom i h1) -> (hindom i h2).
Proof.
Intros.
Inversion_clear H0; Inversion_clear H.
Apply indom_some with n.
Rewrite <- H1; Apply sym_eq; Apply H0.
Save.

Lemma h_subst_disj : (h1,h2,h3:heap)
 (heap_eq h1 h2) -> (hdisj h1 h3) -> (hdisj h2 h3).
Proof.
Intros.
Inversion_clear H0; Inversion_clear H; Apply disj_notboth.
Intro; Cut (hnotindom i h1)\/(hnotindom i h3).
Intro LorR; Case LorR.
Intro L; Left.
Inversion_clear L; Apply notindom_none.
Rewrite <- H; Apply sym_eq; Apply H0.
Intro R; Right; Assumption.
Apply H1.
Save.

Lemma h_subst_append : (h1,h2,h3:heap)
 (heap_eq h1 h2) -> (heap_eq (happend h1 h3) (happend h2 h3)).
Proof.
Intros.
Unfold happend; Inversion_clear H; Apply all_same.
Intro; Replace (h2 i) with (h1 i); Auto.
Save.

Lemma h_subst_append_R : (h1,h2,h3:heap)
 (heap_eq h1 h2) -> (heap_eq (happend h3 h1) (happend h3 h2)).
Proof.
Intros.
Unfold happend; Inversion_clear H; Apply all_same.
Intro; Generalize (h3 i).
Induction o; Auto.
Save.

Lemma disj_emp : (h:heap) (hdisj hempty h).
Proof.
Intros;Apply disj_notboth.
Intro; Left.
Apply notindom_none; Unfold hempty; Trivial.
Save.

Lemma disj_sym : (h1,h2:heap) (hdisj h1 h2) -> (hdisj h2 h1).
Proof.
Intros.
Inversion H; Apply disj_notboth.
Intro.
Cut ((hnotindom i h1)\/(hnotindom i h2)).
Intro OR; Case OR.
Intro LEFT; Right; Assumption.
Intro RIGHT; Left; Assumption.
Apply H0.
Save.

Lemma disj_singleton_free : (a:heap; i:addr; n:int)
 (hdisj (hsingleton i n) (hfree a i)).
Proof.
Intros.
Apply disj_notboth.
Intro.
Cut false=(beq_nat i i0)\/true=(beq_nat i i0).
Intro OR; Case OR.
Intro F; Left.
Apply notindom_none.
Unfold hsingleton.
Rewrite <- F; Trivial.
Intro T; Right.
Apply notindom_none.
Unfold hfree; Rewrite <- T; Trivial.
Generalize (beq_nat i i0); Induction b; Auto.
Save.

Lemma disj_indom_neq : (h1,h2:heap; i1,i2:addr)
 (hdisj h1 h2) -> (hindom i1 h1) -> (hindom i2 h2) -> (~i1=i2).
Proof.
Intros.
Inversion_clear H; Inversion_clear H0; Inversion_clear H1.
Cut ((hnotindom i1 h1)\/(hnotindom i1 h2)); Auto.
Clear H2; Intro.
Intro EQ; Rewrite <- EQ in H0.
Inversion_clear H1.
Inversion_clear H2.
Rewrite H1 in H; Inversion H.
Inversion_clear H2.
Rewrite H1 in H0; Inversion H0.
Save.

Lemma indom_not_notin : (h:heap; i:addr)
 (hindom i h) -> ~(hnotindom i h).
Proof.
Intros.
Intro; Inversion_clear H.
Inversion_clear H0.
Rewrite H1 in H; Inversion H.
Save.

Lemma notin_not_indom : (h:heap; i:addr)
 (hnotindom i h) -> ~(hindom i h).
Proof.
Unfold not; Intros.
Inversion_clear H; Inversion_clear H0.
Rewrite H1 in H; Inversion H.
Save.

Lemma indom_or_notin : (h:heap; i:addr)
 (hindom i h) \/ (hnotindom i h).
Proof.
Intros; Cut (h i)=hvnone\/(EX n:int | (h i)=(hvsome n)).
Intro HH; Inversion HH.
Right; Apply notindom_none; Auto.
Inversion_clear H; Left; Apply indom_some with x; Auto.
Apply opt_dec.
Save.

Lemma free_notin : (h:heap; i:addr)
 (hnotindom i (hfree h i)).
Proof.
Intros.
Apply notindom_none; Unfold hfree.
Replace (beq_nat i i) with true; Auto.
Apply beq_nat_refl.
Save.

Lemma free_preserve_notin : (h:heap; i,j:addr)
 (hnotindom i h) -> (hnotindom i (hfree h j)).
Proof.
Intros.
Inversion H.
Apply notindom_none; Unfold hfree.
Generalize (beq_nat j i); Induction b; Auto.
Save.

Lemma free_preserve_disj : (h1,h2:heap; i:addr)
 (hdisj h1 h2) -> (hdisj (hfree h1 i) h2).
Proof.
Intros.
Inversion_clear H.
Apply disj_notboth; Unfold hfree.
Intro.
Cut ((hnotindom i0 h1)\/(hnotindom i0 h2)).
Intro OR; Case OR.
Intro LEFT; Inversion LEFT; Left.
Apply notindom_none.
Generalize (beq_nat i i0); Induction b; Auto.
Intro Right; Right; Auto.
Auto.
Save.

Lemma notin_free_eq : (h:heap; i:addr)
 (hnotindom i h) -> (heap_eq h (hfree h i)).
Proof.
Intros; Apply all_same.
Intro; Inversion_clear H.
Unfold hfree.
Cut (true=(beq_nat i i0)) \/ (false=(beq_nat i i0)).
Intro TorF; Case TorF.
Intro T; Rewrite <- T.
Replace i0 with i; Auto.
Apply beq_nat_eq; Assumption.
Intro F; Rewrite <- F; Trivial.
Generalize (beq_nat i i0); Induction b; Auto.
Save.

Lemma notin_write_preserve_disj_R : (h1,h2:heap; i:addr; n:hval)
 (hdisj h1 h2) -> (hnotindom i h1) -> (hdisj h1 (hwrite h2 i n)).
Proof.
Intros.
Inversion_clear H.
Apply disj_notboth.
Intro; Cut (hnotindom i0 h1) \/ (hnotindom i0 h2).
Intro aorb; Case aorb.
Intro LEFT; Left; Assumption.
Intro; Cut true=(beq_nat i i0) \/ false=(beq_nat i i0).
Intro EQorNEQ; Case EQorNEQ.
Intro EQ; Left.
Replace i0 with i; Auto.
Apply beq_nat_eq; Assumption.
Intro NEQ; Right.
Apply notindom_none; Unfold hwrite.
Rewrite <- NEQ; Inversion H; Assumption.
Generalize (beq_nat i i0); Induction b; Auto.
Apply H1.
Save.

Lemma notin_write_preserve_disj_L : (h1,h2:heap; i:addr; n:hval)
 (hdisj h1 h2) -> (hnotindom i h2) -> (hdisj (hwrite h1 i n) h2).
Proof.
Intros.
Apply disj_sym.
Apply notin_write_preserve_disj_R.
Apply disj_sym; Assumption.
Assumption.
Save.

Lemma free_read_none : (h:heap; i:addr)
 (hread (hfree h i) i) = hvnone.
Proof.
Intros.
Unfold hread; Unfold hfree.
Replace (beq_nat i i) with true; Auto.
Apply beq_nat_refl.
Save.

Lemma local_free : (h:heap; i,j:addr)
 (~i=j) -> ((hread (hfree h i) j) = (hread h j)).
Proof.
Intros.
Unfold hread; Unfold hfree.
Replace (beq_nat i j) with false; Auto.
Apply beq_nat_neq; Assumption.
Save.

Lemma write_read : (h:heap; i:addr; n:hval)
 (hread (hwrite h i n) i) = (hvsome n).
Proof.
Intros.
Unfold hwrite; Unfold hread.
Replace (beq_nat i i) with true; Auto.
Apply beq_nat_refl.
Save.

Lemma local_write : (h:heap; i,j:addr; n:hval)
 (~i=j) -> ((hread (hwrite h i n) j) = (hread h j)).
Proof.
Intros.
Unfold hwrite; Unfold hread.
Replace (beq_nat i j) with false; Auto.
Apply beq_nat_neq; Assumption.
Save.

Lemma write_free : (h:heap; i:addr; n:hval)
 (heap_eq (hfree (hwrite h i n) i) (hfree h i)).
Proof.
Intros.
Unfold hwrite; Unfold hfree; Apply all_same.
Intro; Cut false=(beq_nat i i0) \/ true=(beq_nat i i0).
Intro ForT; Case ForT.
Intro F; Rewrite <- F; Trivial.
Intro T; Rewrite <- T; Trivial.
Generalize (beq_nat i i0); Induction b; Auto.
Save.

Lemma write_free_sym : (h:heap; i,j:addr; n:hval)
 (~i=j) -> (heap_eq (hfree (hwrite h i n) j) (hwrite (hfree h j) i n)).
Proof.
Intros.
Unfold hwrite; Unfold hfree; Apply all_same.
Intro; Cut true=(beq_nat j i0)\/false=(beq_nat j i0).
Intro TorF; Case TorF.
Intro T; Rewrite <- T.
Replace (beq_nat i i0) with false; Auto.
Replace i0 with j.
Apply beq_nat_neq; Assumption.
Apply beq_nat_eq; Assumption.
Intro F; Rewrite <- F; Trivial.
Generalize (beq_nat j i0); Induction b; Auto.
Save.

Lemma append_emp : (h:heap)
 (heap_eq (happend hempty h) h).
Proof.
Intros.
Unfold hempty; Unfold happend; Apply all_same.
Auto.
Save.

Lemma append_sym : (h1,h2:heap)
 (hdisj h1 h2) ->
 (heap_eq (happend h1 h2) (happend h2 h1)).
Proof.
Intros; Apply all_same.
Intro; Unfold happend.
Inversion_clear H.
Cut (hnotindom i h1) \/ (hnotindom i h2).
Intro OR; Case OR.
Intro LEFT; Inversion_clear LEFT.
Rewrite H.
Generalize (h2 i); Induction o; Trivial.
Intro RIGHT; Inversion_clear RIGHT.
Rewrite H.
Generalize (h1 i); Induction o; Trivial.
Auto.
Save.

Lemma append_assoc_L : (h1,h2,h3:heap)
 (heap_eq (happend (happend h1 h2) h3) (happend h1 (happend h2 h3))).
Proof.
Intros; Apply all_same.
Intro; Unfold happend.
Generalize (h1 i); Induction o; Auto.
Save.

Lemma append_assoc_R : (h1,h2,h3:heap)
 (heap_eq (happend h1 (happend h2 h3)) (happend (happend h1 h2) h3)).
Proof.
Intros; Apply all_same.
Intro; Unfold happend.
Generalize (h1 i); Induction o; Auto.
Save.

Lemma append_singleton_free : (h:heap; i:addr; n:nat)
 ((hread h i) = (hvsome n))
 -> (heap_eq (happend (hsingleton i n) (hfree h i)) h).
Proof.
Intros; Apply all_same.
Intro; Unfold hsingleton; Unfold hfree; Unfold happend.
Cut false=(beq_nat i i0)\/true=(beq_nat i i0).
Intro OR; Case OR.
Intro F; Rewrite <- F; Trivial.
Intro T; Rewrite <- T.
Rewrite <- H; Unfold hread.
Replace i0 with i; Auto.
Apply beq_nat_eq; Auto.
Generalize (beq_nat i i0); Induction b; Auto.
Save.

Lemma append_preserve_notin : (h1,h2:heap; i:addr)
 (hnotindom i h1) -> (hnotindom i h2) -> (hnotindom i (happend h1 h2)).
Proof.
Intros.
Inversion_clear H.
Inversion_clear H0.
Apply notindom_none; Unfold happend.
Rewrite H1; Assumption.
Save.

Lemma append_preserve_indom_L : (h1,h2:heap; i:addr)
 (hindom i h1) -> (hindom i (happend h1 h2)).
Proof.
Intros.
Inversion_clear H.
Apply indom_some with n.
Unfold happend; Rewrite H0; Trivial.
Save.

Lemma append_preserve_indom_R : (h1,h2:heap; i:addr)
 (hindom i h2) -> (hdisj h1 h2) -> (hindom i (happend h1 h2)).
Proof.
Intros.
Inversion_clear H; Inversion_clear H0.
Apply indom_some with n; Unfold happend.
Rewrite H1; Replace (h1 i) with hvnone; Trivial.
Cut (hnotindom i h1).
Intro NN; Inversion_clear NN.
Rewrite H0; Trivial.
Cut (hnotindom i h1) \/ (hnotindom i h2); Auto.
Intro Or; Inversion Or; Auto.
Inversion_clear H0.
Rewrite H1 in H2; Inversion H2.
Save.

Lemma notin_append_R : (h1,h2:heap; i:addr)
 (hnotindom i (happend h1 h2)) -> (hnotindom i h2).
Proof.
Intros.
Inversion_clear H.
Unfold happend in H0.
Apply notindom_none.
Cut ((h1 i) = hvnone)\/(EX n | (h1 i) = (hvsome n)).
Intro OR; Case OR.
Intro eqnone; Rewrite eqnone in H0; Assumption.
Intro eqsome; Inversion eqsome.
Rewrite H in H0.
Inversion H0.
Generalize (h1 i).
Apply opt_dec.
Save.

Lemma notin_append_L : (h1,h2:heap; i:addr)
 (hnotindom i (happend h1 h2)) -> (hnotindom i h1).
Proof.
Intros.
Inversion_clear H.
Unfold happend in H0.
Apply notindom_none.
Cut ((h1 i) = hvnone)\/(EX n | (h1 i) = (hvsome n)).
Intro OR; Case OR.
Intro eqnone; Rewrite eqnone in H0; Assumption.
Intro eqsome; Inversion eqsome.
Rewrite H in H0.
Inversion H0.
Generalize (h1 i).
Apply opt_dec.
Save.

Lemma append_preserve_disj : (h1,h2,h3:heap)
 (hdisj h1 h2) -> (hdisj h1 h3) -> (hdisj h1 (happend h2 h3)).
Proof.
Intros.
Inversion_clear H; Inversion_clear H0; Apply disj_notboth.
Intro; Cut (hnotindom i h1)\/(hnotindom i h2); Auto.
Cut (hnotindom i h1)\/(hnotindom i h3); Auto.
Intros; Clear H1; Clear H.
Inversion H0.
Left; Auto.
Inversion H2.
Left; Auto.
Right; Apply append_preserve_notin; Auto.
Save.

Lemma disj_append_L : (h1,h2,h3:heap)
 (hdisj h1 (happend h2 h3)) -> (hdisj h1 h2).
Proof.
Intros.
Inversion H; Apply disj_notboth.
Intro; Cut (hnotindom i h1)\/(hnotindom i (happend h2 h3)); Auto.
Intro; Clear H0.
Inversion H3.
Left; Auto.
Right; Apply notin_append_L with h3; Auto.
Save.

Lemma disj_append_R : (h1,h2,h3:heap)
 (hdisj h1 (happend h2 h3)) -> (hdisj h1 h3).
Proof.
Intros.
Inversion H; Apply disj_notboth.
Intro; Cut (hnotindom i h1)\/(hnotindom i (happend h2 h3)); Auto.
Intro; Clear H0.
Inversion H3.
Left; Auto.
Right; Apply notin_append_R with h2; Auto.
Save.

Lemma disj_found_notin : (h1,h2:heap; i:addr; n:hval)
 (hdisj h1 h2) -> ((hread h1 i) = (hvsome n)) -> (hnotindom i h2).
Proof.
Intros.
Inversion_clear H.
Cut (hnotindom i h1)\/(hnotindom i h2).
Intro OR; Case OR.
Intro notin1; Inversion_clear notin1.
Unfold hread in H0; Rewrite H0 in H; Inversion H.
Auto.
Auto.
Save.

Lemma disj_indom_notin : (h1,h2:heap; i:addr)
 (hdisj h1 h2) -> (hindom i h1) -> (hnotindom i h2).
Proof.
Intros.
Inversion_clear H0.
Apply disj_found_notin with h1 n; Auto.
Save.

Lemma append_free_sym_L : (h1,h2:heap; i:addr)
 (hnotindom i h2) ->
 (heap_eq (happend (hfree h1 i) h2) (hfree (happend h1 h2) i)).
Proof.
Intros.
Apply all_same.
Intro; Unfold hfree; Unfold happend.
Cut true=(beq_nat i i0) \/ false=(beq_nat i i0).
Intro TorF; Case TorF.
Intro T; Rewrite <- T.
Replace i0 with i.
Inversion_clear H; Assumption.
Apply beq_nat_eq; Assumption.
Intro F; Rewrite <- F; Trivial.
Generalize (beq_nat i i0); Induction b; Auto.
Save.

Lemma append_free_sym_R : (h1,h2:heap; i:addr)
 (hnotindom i h1) ->
 (heap_eq (happend h1 (hfree h2 i)) (hfree (happend h1 h2) i)).
Proof.
Intros.
Apply all_same.
Intro; Unfold hfree; Unfold happend.
Cut true=(beq_nat i i0) \/ false=(beq_nat i i0).
Intro TorF; Case TorF.
Intro T; Rewrite <- T.
Inversion_clear H.
Replace i0 with i.
Rewrite H0; Trivial.
Apply beq_nat_eq; Assumption.
Intro F; Rewrite <- F; Trivial.
Generalize (beq_nat i i0); Induction b; Auto.
Save.

Lemma append_write_sym_L : (h1,h2:heap; i:addr; n:nat)
 (heap_eq (happend (hwrite h1 i n) h2)
   (hwrite (happend h1 h2) i n)).
Proof.
Intros; Apply all_same.
Intro; Unfold hwrite; Unfold happend.
Generalize (beq_nat i i0); Induction b; Auto.
Save.

Lemma append_write_sym_R : (h1,h2:heap; i:addr; n:nat)
 (hnotindom i h1) ->
 (heap_eq (happend h1 (hwrite h2 i n))
   (hwrite (happend h1 h2) i n)).
Proof.
Intros; Apply all_same.
Intro; Unfold hwrite; Unfold happend.
Inversion_clear H.
Cut true=(beq_nat i i0)\/false=(beq_nat i i0).
Intro TorF; Case TorF.
Intro T; Rewrite <- T.
Replace i0 with i.
Rewrite H0; Trivial.
Apply beq_nat_eq; Auto.
Intro F; Rewrite <- F.
Generalize (h1 i0); Induction o; Trivial.
Generalize (beq_nat i i0); Induction b; Auto.
Save.

Lemma append_preserve_read_R : (h1,h2:heap; i:addr)
 (hnotindom i h1) ->
 ((hread (happend h1 h2) i) = (hread h2 i)).
Proof.
Intros.
Unfold hread; Unfold happend.
Inversion_clear H.
Rewrite H0; Trivial.
Save.

Lemma append_preserve_read_L : (h1,h2:heap; i:addr)
 (hnotindom i h2) ->
 ((hread (happend h1 h2) i) = (hread h1 i)).
Proof.
Intros.
Unfold hread; Unfold happend.
Inversion_clear H.
Rewrite H0.
Generalize (h1 i); Induction o; Auto.
Save.

Lemma star_append : (h1,h2,h:heap; i:addr; n:hval)
 ((hread h i) = (hvsome n)) ->
 (heap_eq (happend h1 h2) (hfree h i)) ->
 (heap_eq (happend h1 (hwrite h2 i n)) h).
Proof.
Intros.
Inversion_clear H0; Apply all_same.
Intro; Cut (happend h1 h2 i0)=(hfree h i i0).
Intro.
Unfold happend; Unfold happend in H0.
Cut (h1 i0)=hvnone \/ (EX x | (h1 i0)=(hvsome x)).
Intro OR; Case OR.
Intro eqnone; Rewrite eqnone.
Rewrite eqnone in H0.
Unfold hwrite.
Cut true=(beq_nat i i0) \/ false=(beq_nat i i0).
Intro TorF; Case TorF.
Intro T; Rewrite <- T.
Rewrite <- H.
Unfold hread; Replace i0 with i; Auto.
Apply beq_nat_eq; Assumption.
Intro F; Rewrite <- F.
Rewrite H0.
Unfold hfree; Rewrite <- F; Trivial.
Generalize (beq_nat i i0); Induction b; Auto.
Intro eqsome; Inversion_clear eqsome.
Rewrite H2; Rewrite H2 in H0.
Rewrite H0.
Cut (~i=i0).
Intro neq; Unfold hfree.
Replace (beq_nat i i0) with false; Trivial.
Apply beq_nat_neq; Assumption.
Intro eqf; Rewrite eqf in H0.
Unfold hfree in H0.
Cut true = (beq_nat i0 i0).
Intro T; Rewrite <- T in H0; Inversion H0.
Apply beq_nat_refl.
Apply opt_dec.
Auto.
Save.

(******************** Add to "fsl" hints database *******************)
Hints Resolve beq_nat_neq : fsl.
Hints Immediate disj_emp append_preserve_notin : fsl.
Hint append_preserv_read_R_tac1 : fsl 
  := Extern 1 (hread (happend ?1 ?2) ?3)=(Some nat ?4)
     Match Context With
        [rw:(hread ?2 ?3)=(Some nat ?4) 
              |- (hread (happend ?1 ?2) ?3) = (Some nat ?4)] 
	 -> 
             Rewrite <- rw; Apply append_preserve_read_R; 
             Apply disj_found_notin with ?2 ?4; 
             [Apply disj_sym; Assumption | Assumption].
